# Desactiva el bytecode de Python
import sys
sys.dont_write_bytecode = True

# Kafka
import kafka

# Librerías
import time
import atexit
import pandas as pd
from typing import List

# SQLAlchemy
import sqlalchemy as sa

# Demonio
from demonio import Demonio

# Fecha
from fecha import Fecha

# Funciones
import general

# Modelos
import modelos
from modelos import RegistroLocal, RegistroKafka

# --------------------------- Ajustes del Consumidor --------------------------

# ----------- Datos SQLite ------------
SQ_ARCHIVO = 'datos.db'

# --------- Datos PostgreSQL ----------
PG_HOST       = 'base.datos.com'
PG_PUERTO     = 5433
PG_USUARIO    = 'usuario'
PG_CLAVE      = 'clave'
PG_BASE_DATOS = 'epi_puma'

# ------------ Datos Kafka ------------
KAFKA_BOOTSTRAP_SERVER = 'servidor.kafka:9092'
KAFKA_SCHEMA_REGISTRY  = 'http://servidor.kafka:8081'
KAFKA_GROUP_ID         = 'covid19_epi_puma_30d'
KAFKA_TOPIC            = 'covid19'
KAFKA_PARTITION        = 0

# Archivo de la fecha local
ARCHIVO_FECHA = 'fecha.txt'

# Fecha inicial de la base de datos remota
FECHA_INICIAL = '2021-03-01'

# Tiempo sin recibir mensajes para considerar que hubo un cambio de día (segundos)
TIEMPO_PURGA_REGISTROS = 60

# Los registros se guardarán en paquetes de este tamaño (cantidad de registros)
LIMITE_CARGA_REGISTROS = 100000

# ----------------------------- Variables Globales ----------------------------

# Fecha local
FECHA_LOCAL = Fecha(ARCHIVO_FECHA, pd.Timestamp(FECHA_INICIAL))

# Conexiones
CONEXION_KAFKA:  kafka.Conexion             = None
CONEXION_LOCAL:  modelos.ConexionSQLite     = None
CONEXION_REMOTA: modelos.ConexionPostgreSQL = None

# Variables globales
CAMBIO_DE_DIA                   = False
BASE_ACTUALIZADA                = False
TRANSACCION_FINALIZADA          = False
TIEMPO_ANTERIOR: float          = None
DIA_ACTUAL: pd.Timestamp        = None
DIA_ACTUALIZACION: pd.Timestamp = None
INDICE_KAFKA: int               = None

# Variables globales de la carga de registros
INSERTADOS:   List[RegistroKafka] = []
ELIMINADOS:   List[RegistroKafka] = []
ACTUALIZADOS: List[RegistroKafka] = []
PENDIENTES:   List[RegistroKafka] = []

# -------------------------- Funciones de inserción ---------------------------

def insertar_registros_locales(insertados: pd.DataFrame):
	''' Inserta los registros individuales de la lista INSERTADOS en la base de datos local '''

	if insertados.empty:
		return

	# Se omiten los registros con un grupo de edad inválido
	insertados.dropna(subset = ['GRUPO_EDAD'], inplace = True)

	# Se filtran los registros positivos y no positivos
	positivos = insertados[insertados['RESULTADO'] == 'CONFIRMADO']
	no_positivos = insertados[insertados['RESULTADO'] != 'CONFIRMADO']

	# Se filtran los registros positivos que tengan una antigüedad mayor a 30 días
	inicio = FECHA_LOCAL.actual - pd.Timedelta(days = 30)
	positivos = positivos[(positivos['FECHA_SINTOMAS'] >= inicio) | (positivos['FECHA_DEF'] >= inicio)]

	# Se eliminan todos los registros que no sean positivos
	ids_eliminados = no_positivos['ID_REGISTRO'].to_list()
	eliminar: sa.sql.Delete = sa.delete(RegistroLocal)
	CONEXION_LOCAL.sesion.execute(eliminar.where(RegistroLocal.ID_REGISTRO.in_(ids_eliminados)))
	CONEXION_LOCAL.sesion.flush()

	# Se insertan unicamente los registros positivos
	if not positivos.empty:
		CONEXION_LOCAL.sesion.execute(sa.insert(RegistroLocal), positivos.to_dict(orient = 'records'))
		CONEXION_LOCAL.sesion.flush()

def eliminar_registros_locales(eliminados: pd.DataFrame):
	''' Elimina los registros individuales de la lista ELIMINADOS de la base de datos local '''

	if eliminados.empty:
		return

	# Se eliminan todos los registros dados
	ids_eliminados = eliminados['ID_REGISTRO'].to_list()
	eliminar: sa.sql.Delete = sa.delete(RegistroLocal)
	CONEXION_LOCAL.sesion.execute(eliminar.where(RegistroLocal.ID_REGISTRO.in_(ids_eliminados)))
	CONEXION_LOCAL.sesion.flush()

def eliminar_registros_anteriores():
	''' Elimina los registros de la base local que tengan una antigüedad mayor a 30 días '''

	# Se obtiene la fecha de inicio
	inicio = FECHA_LOCAL.actual - pd.Timedelta(days = 30)

	# Se eliminan los registros sin fecha de defunción con una antigüedad mayor a 30 días
	CONEXION_LOCAL.sesion.query(RegistroLocal) \
	.filter(RegistroLocal.FECHA_DEF == None) \
	.filter(RegistroLocal.FECHA_SINTOMAS < inicio).delete()

	# Se eliminan los registros con fecha de defunción con ambas condiciones verdaderas:
	# - Una antigüedad mayor a 30 días
	# - Una fecha de defunción mayor a 30 días
	CONEXION_LOCAL.sesion.query(RegistroLocal) \
	.filter(RegistroLocal.FECHA_DEF != None) \
	.filter(
		(RegistroLocal.FECHA_SINTOMAS < inicio) &
		(RegistroLocal.FECHA_DEF < inicio)
	).delete()

def insertar_registros_remotos():
	'''
		Carga los datos almacenados localmente en la base de datos remota
		- La base de datos local almacena cada registro confirmado de manera individual
		- La base de datos remota almacena los registros por fecha, municipio y grupo de edad
	'''

	# Se cargan los registros almacenados localmente
	registros = pd.read_sql(CONEXION_LOCAL.sesion.query(RegistroLocal).statement, CONEXION_LOCAL.enlace)

	# Se convierten las fechas a formato datetime
	registros['FECHA_SINTOMAS'] = pd.to_datetime(registros['FECHA_SINTOMAS'])
	registros['FECHA_DEF'] = pd.to_datetime(registros['FECHA_DEF'], format = '%Y-%m-%d', errors = 'coerce')

	# Se guardan los registros en la base de datos remota
	general.cargar_valores_reales(CONEXION_REMOTA, registros, FECHA_LOCAL.actual)

# ------------------------- Funciones del consumidor --------------------------

def cantidad_registros_pendientes():
	''' Regresa la cantidad de registros pendientes por cargar en la base de datos '''

	return len(INSERTADOS) + len(ELIMINADOS) + len(ACTUALIZADOS) + len(PENDIENTES)

def refrescar_dia_actual(fecha_registro: pd.Timestamp, purgar = False):
	'''
		Refresca la fecha actual según la fecha de un registro y determina si dicho registro
		corresponde a un dia diferente al actual, en cuyo caso se toma como un cambio de dia
	'''

	global DIA_ACTUAL, DIA_ACTUALIZACION, CAMBIO_DE_DIA

	# Si la fecha del registro es mayor a la fecha actual, se toma como un cambio de dia
	if fecha_registro > pd.Timestamp(DIA_ACTUAL):
		if purgar:
			print('Purgando registros pendientes...\n')
		DIA_ACTUALIZACION = DIA_ACTUAL
		DIA_ACTUAL = fecha_registro
		CAMBIO_DE_DIA = True

	# Si el dia actual es nulo, se asigna la fecha del registro
	elif DIA_ACTUAL is None:
		DIA_ACTUAL = fecha_registro

def enlistar_registro(registro: RegistroKafka):
	''' Enlista el registro en la lista correspondiente para su posterior procesamiento '''

	# Si se detecta un cambio de dia, se continúa con el siguiente ciclo y se guarda
	# el ultimo registro como pendiente para procesarlo en el siguiente ciclo
	if CAMBIO_DE_DIA:
		PENDIENTES.append(registro)
	elif registro.TYPE == 'INSERT':
		INSERTADOS.append(registro)
	elif registro.TYPE == 'DELETE':
		ELIMINADOS.append(registro)
	elif registro.TYPE == 'UPDATE':
		ACTUALIZADOS.append(registro)

def inicio():
	''' Función de inicialización del consumidor de Kafka '''

	global CONEXION_LOCAL, CONEXION_REMOTA, CONEXION_KAFKA, TIEMPO_ANTERIOR

	# -------------------------- Finalización del script --------------------------

	def finalizar_proceso():
		''' Finaliza el script y cierra las conexiones a la base de datos y kafka '''

		print('\nFinalizando proceso...')

		# Si la transacción no fue finalizada o hay algún registro pendiente por
		# cargar en la base de datos, se regresa el indice de kafka al inicio
		if not TRANSACCION_FINALIZADA or cantidad_registros_pendientes() > 0:
			CONEXION_KAFKA.regresar_indice_remoto()
			print('Indice de Kafka restaurado')

		# Se cierran las conexiones de las bases de datos y kafka
		CONEXION_LOCAL.cerrar()
		CONEXION_REMOTA.cerrar()
		CONEXION_KAFKA.cerrar()

		print('Conexiones finalizadas')

	atexit.register(finalizar_proceso)

	# ------------------------- Conexión a la base local --------------------------

	print('Conectando a la base de datos local...')

	CONEXION_LOCAL = modelos.ConexionSQLite \
	(
		SQ_ARCHIVO
	)

	print('  Base de datos conectada\n')

	# ------------------------ Conexión a la base remota --------------------------

	print('Conectando a la base de datos remota...')

	CONEXION_REMOTA = modelos.ConexionPostgreSQL \
	(
		PG_HOST,
		PG_PUERTO,
		PG_USUARIO,
		PG_CLAVE,
		PG_BASE_DATOS
	)

	print('  Base de datos conectada\n')

	# --------------------------- Conexión a Kafka -------------------------------

	print('Conectando a Kafka...')

	CONEXION_KAFKA = kafka.Conexion \
	(
		KAFKA_BOOTSTRAP_SERVER,
		KAFKA_GROUP_ID,
		KAFKA_SCHEMA_REGISTRY,
		KAFKA_TOPIC,
		KAFKA_PARTITION,
		RegistroKafka.desde_diccionario
	)

	print('  Kafka conectado\n')

	# Se inicializa la base de datos remota
	modelos.inicializar_base_remota(CONEXION_REMOTA)

	# Se obtiene el tiempo inicial
	TIEMPO_ANTERIOR = time.time()

def ciclo():
	''' Ciclo principal del consumidor de Kafka '''

	global CAMBIO_DE_DIA, TRANSACCION_FINALIZADA, TIEMPO_ANTERIOR

	# Si se detecta un cambio de dia o se alcanza el limite de registros
	# cargados, se procesan los registros cargados hasta el momento
	if CAMBIO_DE_DIA or cantidad_registros_pendientes() >= LIMITE_CARGA_REGISTROS:

		# Se marca la transacción como no finalizada
		TRANSACCION_FINALIZADA = False

		print(f'Procesando {len(INSERTADOS)} registros a insertar...')
		print(f'Procesando {len(ELIMINADOS)} registros a eliminar...')
		print(f'Procesando {len(ACTUALIZADOS)} registros a actualizar...')

		# Se convierten los registros de Kafka en DataFrames
		insertados = general.generar_dataframe(INSERTADOS)
		eliminados = general.generar_dataframe(ELIMINADOS)
		actualizados = general.generar_dataframe(ACTUALIZADOS)

		print('Realizando cambios en la base de datos local...')

		# Se eliminan los registros de la lista de 'ELIMINADOS' y 'ACTUALIZADOS'
		eliminar_registros_locales(eliminados)
		eliminar_registros_locales(actualizados)

		# Se insertan los registros de la lista de 'INSERTADOS' y 'ACTUALIZADOS'
		insertar_registros_locales(insertados)
		insertar_registros_locales(actualizados)

		print('Persistiendo los cambios en la base de datos local...')

		# Se hacen efectivos los cambios en la base de datos local
		CONEXION_LOCAL.sesion.commit()
		CONEXION_LOCAL.limpiar()

		# Se marca la transacción como finalizada
		TRANSACCION_FINALIZADA = True

		# Se recorre el indice inicial de kafka al indice del ultimo registro procesado
		CONEXION_KAFKA.avanzar_indice_local()

		print('    Base de datos local actualizada!\n')

		# La base de datos remota se actualiza unicamente si se detecta un cambio de dia
		if CAMBIO_DE_DIA:
			print('Cambio de dia detectado...\n')
			CAMBIO_DE_DIA = False

			# Se actualiza la base de datos remota hasta que se alcance el día de actualización
			while FECHA_LOCAL < DIA_ACTUALIZACION:
				print('Día local: ', FECHA_LOCAL.actual.date())
				print('Actualizando la base de datos local y remota...')

				# Se actualizan los datos locales y remotos
				eliminar_registros_anteriores()
				insertar_registros_remotos()

				print('Persistiendo los cambios en las bases de datos...')

				# Se hacen efectivos los cambios en las bases de datos
				CONEXION_REMOTA.sesion.commit()
				CONEXION_LOCAL.sesion.commit()
				CONEXION_LOCAL.limpiar()

				print('    Base de datos remota actualizada!\n')

				# Se avanza la fecha local
				FECHA_LOCAL.avanzar()

			print(f'Base de datos remota al día! ({FECHA_LOCAL.actual.date()})')

		# Se limpian las listas de registros
		INSERTADOS.clear()
		ELIMINADOS.clear()
		ACTUALIZADOS.clear()

	# Si no hay registros pendientes, se obtiene el siguiente registro
	if len(PENDIENTES) == 0:

		# Se obtiene el siguiente registro y el tiempo actual
		registro: RegistroKafka = CONEXION_KAFKA.obtener_registro()
		tiempo_actual = time.time()

		# Si el registro no es nulo, se refresca el tiempo anterior
		if registro is not None:
			TIEMPO_ANTERIOR = tiempo_actual

		# Si transcurre cierto tiempo sin recibir mensajes, se considera que hubo
		# un cambio de dia y se cargan los datos leídos hasta el momento
		elif tiempo_actual - TIEMPO_ANTERIOR >= TIEMPO_PURGA_REGISTROS:

			# Solo se refresca el dia actual si hay registros cargados en las listas
			if cantidad_registros_pendientes() > 0:
				refrescar_dia_actual(DIA_ACTUAL + pd.Timedelta(days = 1), purgar = True)

	# En caso contrario, se obtiene el siguiente registro de la lista de pendientes
	else:
		registro = PENDIENTES.pop(0)

	# Si el registro es nulo, se continúa con el siguiente ciclo
	if registro is None:
		return

	# Se obtiene la fecha del registro
	fecha_registro = registro.fecha_actualizacion

	# La fecha de actualización de los registros eliminados es la fecha del archivo
	# anterior, por lo que se incrementa en un dia para que coincida con el dia actual
	if registro.TYPE == 'DELETE':
		fecha_registro += pd.Timedelta(days = 1)

	# Se refresca el dia actual
	refrescar_dia_actual(fecha_registro)

	# Se enlista el registro en la lista correspondiente
	enlistar_registro(registro)

# Se muestra la interfaz de consola del demonio
Demonio(inicio, ciclo, 'consumidor.pid', 'consumidor.log').interfaz()